"""Implements process for process manager."""
import multiprocessing as mp
import sys
from time import monotonic as current_time
from typing import Callable, Optional

from sanic.exceptions import RequestTimeout, ServiceUnavailable
from sanic.http import Stage
from sanic.log import error_logger, logger
from sanic.server.protocols.base_protocol import SanicProtocol
from sanic.server.protocols.http_protocol import HttpProtocol

from .reloaders import SanicReloader


def _close(self, timeout: float | None = None):  # pylint: disable-msg=unused-argument
    """
    Monkey patch https://github.com/sanic-org/sanic/blob/main/sanic/server/protocols/base_protocol.py#L94
    See https://jira.visionlabs.ru/browse/LUNA-6299 and sanic discord for details
    """
    if self.transport:
        self.transport.close()
        self.abort()


SanicProtocol.close = _close


def check_timeouts(self):
    """
    Runs itself periodically to enforce any expired timeouts.
    """
    try:
        if not self._task:
            return
        duration = current_time() - self._time
        stage = self._http.stage
        if stage is Stage.IDLE and duration > self.keep_alive_timeout:
            logger.debug("KeepAlive Timeout. Closing connection.")
        elif stage is Stage.REQUEST and duration > self.request_timeout:
            logger.debug("Request Timeout. Closing connection.")
            self._http.exception = RequestTimeout("Request Timeout")
        elif stage is Stage.HANDLER and self._http.upgrade_websocket:
            logger.debug("Handling websocket. Timeouts disabled.")
            return
        elif stage in (Stage.HANDLER, Stage.RESPONSE, Stage.FAILED) and duration > self.response_timeout:
            logger.debug("Response Timeout. Closing connection.")
            self._http.exception = ServiceUnavailable("Response Timeout")
        else:
            interval = (
                min(
                    self.keep_alive_timeout,
                    self.request_timeout,
                    self.response_timeout,
                )
                / 2
            )
            self._handler_task = self.loop.call_later(max(0.1, interval), self.check_timeouts)
            return
        cancel_msg_args = ()
        if sys.version_info >= (3, 9):
            cancel_msg_args = ("Cancel connection task with a timeout",)
        self._task.cancel(*cancel_msg_args)
    except Exception:  # pylint: disable-msg=broad-exception-caught
        error_logger.exception("protocol.check_timeouts")


HttpProtocol.check_timeouts = check_timeouts


def close(self, timeout: Optional[float] = None):  # pylint: disable-msg=missing-function-docstring,unused-argument
    if self._handler_task:
        self._handler_task.cancel()
    if self.transport:
        self.transport.close()
        self.abort()


HttpProtocol.close = close


class Process:
    """
    Wrapper around multiprocessing.Process that allows to restart wrapped process.

    Attributes:
        target: target for mp.Process
        name: process name
        process: mp.Process
        restarting: A bool, whether process is restarting. Used to distinguish failure from expected exit.
    """

    def __init__(self, name: str, target: Callable):
        """
        Args:
            name: name of the process
            target: target function for mp.Process
        """

        self.name = name
        self.target: Callable = target
        self.process: Optional[mp.Process] = None
        self.restarting = False

    def start(self, restart=False):
        """
        Start or restart the process.

        Sets restarting to True during the restart, so the manager can distinguish between
        failure and restart.

        Args:
            restart: A bool, whether to restart process if its running.
        """
        if self.process is not None and not restart:
            return

        if self.process:
            self.restarting = True
            self.stop()

        process = self.makeProcess()
        process.start()
        self.process = process
        self.restarting = False

    def stop(self):
        """Stop the process."""
        self.process.terminate()
        self.process.join()

    def restart(self):
        """Restart the process."""
        self.start(restart=True)

    def isAlive(self) -> bool:
        """Check whether the process is alive."""
        return self.process.is_alive()

    def makeProcess(self) -> mp.Process:
        """Configure and return Process instance."""
        ctx = mp.get_context("spawn")
        return ctx.Process(target=self.target, name=self.name)


class SanicProcess(Process):
    """Generic class for sanic web application process."""

    sock = None

    def getReloader(self, config) -> SanicReloader:
        """Fabric for reloader instance for sanic application process."""
        return SanicReloader(self, config)

    def makeProcess(self) -> mp.Process:
        """Configure and return Process instance."""
        ctx = mp.get_context("spawn")

        if self.sock:
            return ctx.Process(target=self.target, name=self.name, args=(self.sock,))

        return ctx.Process(target=self.target, name=self.name)
