"""Implements base class for RPC client."""

from typing import Any, Awaitable, List

from ..utils.log import Logger
from .transport.zmq import ZMQTransport


class Client:
    """
    Base class for RPC clients.

    Attributes:
        transport: transport
    """

    def __init__(
        self,
        address: str,
        logger: Logger,
        connectTimeout: int = 20000,
        requestTimeout: int = 60000,
        responseTimeout: int = 60000,
    ):
        """
        Args:
            address: rpc server address
            logger: initialized logger instance
            connectTimeout: connect timeout
            requestTimeout: maximum time before a receive operation
            responseTimeout: maximum time before a send operation
        """
        self.transport = ZMQTransport(
            address,
            logger,
            connectTimeout=connectTimeout,
            requestTimeout=requestTimeout,
            responseTimeout=responseTimeout,
        )

    def close(self):
        """Close client."""
        self.transport.close()

    def call(self, command: str, args: List[Any], requestId: str) -> Awaitable:
        """
        Call using transport.

        Args:
            command: A str, used to distinguish commands
            args: List of positional arguments
            requestId: A str, request id

        Returns:
            coroutine
        """
        return self.transport.call(command, args, requestId)
