""" Module release universal engine to inspect database. """
from contextlib import asynccontextmanager
from typing import Any, List, Tuple, Union

from ..adaptors.adaptor import DBAdaptor
from ..enums.databases import Databases
from .healthcheck import inspectAlembicRevision
from .log import Logger


class UniAdaptorEngine:
    """
    Universal engine class. Using adaptor for db connections.

    Attributes:
        _dbAdaptor: database adaptor
        _logger: logger for adaptor
    """

    def __init__(self, dbSetting: "DBSetting"):
        """
        Args:
            dbSetting: database settings
        """
        self._dbAdaptor = DBAdaptor(
            dbSetting.type,
            dbSetting.dsn,
            dbSetting.userName,
            dbSetting.password,
            sessionCount=1,
        )
        self._logger = Logger("Uni engine logger")

    @asynccontextmanager
    async def adaptorContext(self):
        """
        Context manager for connection to database using adaptor.
        """
        await self._dbAdaptor.initiateAdaptor()
        try:
            yield self
        finally:
            await self._dbAdaptor.close()

    async def inspectDBSchema(self) -> List[str]:
        """
        Inspect database schema
        Returns:
            list of current database tables and views
        """
        from sqlalchemy import text  # pylint: disable-msg=C0415
        from sqlalchemy.sql import sqltypes  # pylint: disable-msg=C0415

        tableNameSelectMap = {
            Databases.POSTGRES.value: text(
                "SELECT c.relname FROM pg_class c "
                "JOIN pg_namespace n ON n.oid = c.relnamespace "
                "WHERE n.nspname NOT IN ('pg_catalog', 'information_schema') AND c.relkind in ('r', 'p')"
            ).columns(relname=sqltypes.Unicode),
            Databases.ORACLE.value: text(
                f"SELECT LOWER(table_name) FROM all_tables WHERE nvl(tablespace_name, 'no tablespace')"
                f"NOT IN ('SYSTEM', 'SYSAUX') AND OWNER = '{self._dbAdaptor.dblogin.upper()}' AND "
                f"IOT_NAME IS NULL AND DURATION IS NULL"
            ),
        }
        viewNameSelectMap = {
            Databases.POSTGRES.value: text(
                "SELECT c.relname FROM pg_class c "
                "JOIN pg_namespace n ON n.oid = c.relnamespace "
                "WHERE n.nspname NOT IN ('pg_catalog', 'information_schema') AND c.relkind IN ('v', 'm')"
            ),
            # oracle return view names from 'table names' query
        }

        async with self._dbAdaptor.connection(self._logger) as connection:
            dbReply = await connection.fetchall(tableNameSelectMap[self._dbAdaptor.dbtype])
            tableNames = [row[0] for row in dbReply]

            isNeedViewSeparateQuery = self._dbAdaptor.dbtype in viewNameSelectMap
            if isNeedViewSeparateQuery:
                dbReply = await connection.fetchall(viewNameSelectMap[self._dbAdaptor.dbtype])
                tableNames.extend([row[0] for row in dbReply])

        return tableNames

    async def execute(self, querySt: Union["sqlalchemy.sql.selectable.Select", str]) -> Any:
        """
        Execute database query
        Args:
            querySt: query statement
        Returns:
            query result
        """
        async with self._dbAdaptor.connection(self._logger) as connection:
            return await connection.fetchall(querySt)

    async def inspectAlembicRevision(self) -> Tuple[str, str]:
        """
        Inspect alembic revisions. Get a real revision version from the database and an expected revision version
        from alembic versions files.
        Returns:
            tuple(<real migration version>, <expected migration version>)
        """
        return await inspectAlembicRevision(adaptor=self._dbAdaptor)
